# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib
import logging
import functools
import memcache
from tornado.escape import utf8
from tornado.options import define, options

define("cache_key_prefix", "", str, "cache key prefix to avoid key conflict")
define("cache_key_prefix_another", "", str,
       "another cache in different cache prefix, should be delete at the same time")
define("cache_enabled", True, bool, "whether cache is enabled")
manager = None


def cache(key=None, timeout=3600, args_as_key=True):
    def _wrapper(func):
        @functools.wraps(func)
        def __wrapper(*args, **kw):
            if not options.cache_enabled:
                return func(*args, **kw)
            _key = key_gen(key, func, args_as_key, *args, **kw)
            value = manager.get(_key)
            if not value:
                value = func(*args, **kw)
                if value:
                    manager.set(_key, value, timeout)

            return value
        return __wrapper

    return _wrapper


def delete(key):
    key_ = key
    if options.cache_key_prefix:
        key = "%s:%s" % (options.cache_key_prefix, key_)
    manager.delete(key)

    if options.cache_key_prefix_another:
        another_key = "%s:%s" % (options.cache_key_prefix_another, key_)
        if another_key != key:
            manager.delete(another_key)


def key_gen(key="", func=None, args_as_key=True, *args, **kw):
    assert key or func, "key and func must has one"
    if key:
        if args_as_key and "%s" in key:
            args_ = []
            for arg in args:
                if isinstance(arg, (basestring, unicode, int, long)):
                    args_.append(arg)
            key = key % tuple(args_)
    else:
        code = hashlib.md5()
        code.update("%s-%s" % (func.__module__, func.__name__))

        # copy args to avoid sort original args
        c = list(args[:])
        # sort c to avoid generate different key when args is the same
        # but sequence is different
        c.sort()
        c = [utf8(v) if isinstance(v, (basestring, unicode))
             else str(v) for v in c]
        code.update("".join(c))

        c = ["%s=%s" % (k, v) for k, v in kw.iteritems()]
        c.sort()
        code.update("".join(c))

        key = code.hexdigest()

    if options.cache_key_prefix:
        key = "%s:%s" % (options.cache_key_prefix, key)

    return key


def setup(servers, timeout=3):
    global manager

    if manager is None:
        manager = CacheManager(servers, timeout)
    return manager


def reconnect(func):
    @functools.wraps(func)
    def _wrapper(self, *args, **kwargs):
        try:
            r = func(self, *args, **kwargs)

            return r
        except Exception:
            logging.exception("memcache server closed!")
            self.close()

    return _wrapper


class CacheManager(object):

    def __init__(self, servers, timeout=3):
        self.servers = servers
        self.default_timeout = int(timeout)
        self._cache = memcache.Client(self.servers)

        logging.debug("Memcached start client %s" % servers)

    @property
    def cache(self):
        if self._cache is None:
            self._cache = memcache.Client(self.servers)

        return self._cache

    @reconnect
    def add(self, key, value, timeout=0):
        if isinstance(value, unicode):
            value = utf8(value)

        return self.cache.add(utf8(key), value,
                              timeout or self.default_timeout)

    @reconnect
    def get(self, key, default=None):
        val = self.cache.get(utf8(key))
        if val is None:
            return default
        else:
            if isinstance(val, basestring):
                return utf8(val)
            else:
                return val

    @reconnect
    def set(self, key, value, timeout=0):
        if isinstance(value, unicode):
            value = utf8(value)
        return self.cache.set(utf8(key), value,
                              timeout or self.default_timeout)

    @reconnect
    def delete(self, key):
        return self.cache.delete(utf8(key))

    @reconnect
    def get_many(self, keys):
        return self.cache.get_multi(map(utf8, keys))

    def close(self, **kwargs):
        try:
            self._cache.disconnect_all()
        except Exception:
            self._cache = None

    @reconnect
    def stats(self):
        return self.cache.get_stats()

    @reconnect
    def flush_all(self):
        self.cache.flush_all()


if __name__ == "__main__":
    setup(["127.0.0.1"])

    @cache()
    def test(a):
        return a

    a = test(1)
    key = key_gen("", test, 1)
    assert manager.get(key) == a



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Utility library"""

import re
import random
import importlib
from tornado.web import url
from tornado.options import options
from tornado import escape
from torndb import Row

from djinn.errors import TemplateContextError

ALPHABET_FULL = '23456789abcdefghijkmnpqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ'
ALPHABET_LOWER_ONLY = '23456789abcdefghijkmnpqrstuvwxyz'


class Context(dict):

    """Template context container.

    A container which will return empty string silently if the key is not exist
    rather than raise AttributeError when get a item's value.

    It will raise TemplateContextError if debug is True and the key
    does not exist.

    The context item also can be accessed through get attribute.

    """

    def __setattr__(self, key, value):
        self[key] = value

    def __str__(self):
        return str(self)

    def __iter__(self):
        return iter(self.items())

    def __getattr__(self, key):
        """Get a context attribute.

        Raise TemplateContextError if the attribute not be set to
        avoid confused AttributeError exception when debug is True, or return ""

        """
        if key in self:
            return self[key]
        elif options.debug:
            raise TemplateContextError("'%s' does not exist in context" % key)
        else:
            return ""

    def __hasattr__(self, key):
        if key in self:
            return True
        else:
            return False


def load_url_handlers(handlers_root_module, handlers_modules,
                      prefix=""):
    """Load URL handlers

    ``handlers_root_module``: The root module of all of handlers
    ``handlers_modules``: A list of module which is the container of a
    suite of handlers
    ``prefix``: URL prefix
    """
    handlers = []
    for handlers_module in handlers_modules:
        module = importlib.import_module(".%s" %
                                         handlers_module, handlers_root_module)
        module_hanlders = getattr(module, "handlers", None)
        if module_hanlders:
            _handlers = []
            for handler in module_hanlders:
                try:
                    patten = r"%s%s" % (prefix, handler[0])
                    if len(handler) == 2:
                        _handlers.append((patten,
                                          handler[1]))
                    elif len(handler) == 3:
                        _handlers.append(url(patten,
                                             handler[1],
                                             name=handler[2])
                                         )
                    else:
                        pass
                except IndexError:
                    pass

            handlers.extend(_handlers)

    return handlers


def get_count(row):
    """Get count from a query result

    ``row``: Row object of a torndb query result
    """
    if row:
        count = row.c  # c as count by name convention
    else:
        count = 0

    return count


def get_column_values(rows, column=None):
    """Get all values of a column from a query result

    ``rows``: A list of Row object of a torndb query result
    ``column``: Column name which value will be collected
    """
    result = []
    if rows:
        for row in rows:
            if column:
                result.append(row.get(column, None))
            else:
                result.append(row.values()[0])

    return result


def pick_row_attrs(row, attrs):
    if not row or not isinstance(row, dict):
        return row

    row_ = Row()
    for name in attrs:
        row_[name] = row.get(name, None)

    return row_


def gen_uid(size=6, alphabet=ALPHABET_LOWER_ONLY):
    s = []
    # random part of uid
    for i in xrange(0, size):
        s.append(random.choice(alphabet))

    return "".join(s)


def dict_value_to_string(data):
    assert isinstance(data, dict), "data should be a dict"

    for k, v in data.iteritems():
        if isinstance(v, unicode):
            data[k] = escape.utf8(v)
        elif isinstance(v, dict):
            data[k] = escape.json_encode(v)
        else:
            data[k] = v

    return data


def parse_on_off(status, default="on"):
    if not status:
        return default

    return "on" if status == "on" else "off"


def shorten_content(content, length):
    content = content.strip(" \n\t")
    content = content if len(
        content) <= length else "%s..." % content[0:length - 3]
    return content


def compare_version(v_a, v_b):
    """compare version numbers

    normal cases:
    1.10.2 > 1.1.3
    1.2.1 < 1.2.2
    1.2.0000 = 1.2.0

    exception cases:
    1.2.1abc > 1.2.1
    None < 1.0
    "  " < 1.0
    """
    def normalize(v):
        vv = []
        v = escape.utf8(v)

        if v:
            for x in re.sub(r'(\.0+)*$', '', v).split("."):
                try:
                    x = int(x)  # 001 to 1
                    vv.append(x)
                except ValueError:
                    # may have characters in the tail sometimes
                    vv.append(x)

        return vv

    return cmp(normalize(v_a), normalize(v_b))



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
import traceback
import logging

from tornado import escape
from tornado.options import options
from tornado.web import RequestHandler as BaseRequestHandler, HTTPError
from djinn import errors
from djinn.utils import Context

REMOVE_SLASH_RE = re.compile(".+/$")


class BaseHandler(BaseRequestHandler):

    def prepare(self):
        self.remove_slash()
        self.prepare_context()
        self.traffic_threshold()

    def traffic_threshold(self):
        pass

    def prepare_context(self):
        self._context = Context()

    def remove_slash(self):
        if self.request.method == "GET":
            if REMOVE_SLASH_RE.match(self.request.path):
                # remove trail slash in path
                uri = self.request.path.rstrip("/")
                if self.request.query:
                    uri += "?" + self.request.query

                self.redirect(uri)

    def render_string(self, template_name, **kwargs):
        """Override default render_string, add context to template."""
        assert "context" not in kwargs, "context is a reserved word for \
                template context valuable."
        kwargs['context'] = self._context
        kwargs['url_escape'] = escape.url_escape

        return super(BaseHandler, self).render_string(template_name, **kwargs)

    def get_argument(self, name, default=BaseRequestHandler._ARG_DEFAULT, strip=True):
        value = super(BaseHandler, self).get_argument(name, default, strip)

        return escape.utf8(value) if isinstance(value, unicode) else value

    def get_int_argument(self, name, default=0):
        try:
            return int(self.get_argument(name, default))
        except ValueError:
            return default


class APIHandler(BaseHandler):

    def finish(self, chunk=None, message=None):
        if chunk is None:
            chunk = {}

        if isinstance(chunk, dict):
            chunk = {"meta": {"code": errors.HTTP_OK}, "data": chunk}

            if message:
                chunk["message"] = message
        elif isinstance(chunk, errors.HTTPAPIError):
            chunk = str(chunk)

        callback = escape.utf8(self.get_argument("callback", None))
        if callback:
            self.set_header("Content-Type", "application/x-javascript")

            if isinstance(chunk, dict):
                chunk = escape.json_encode(chunk)

            self._write_buffer = [callback, "(", chunk, ")"] if chunk else []
            super(APIHandler, self).finish()
        else:
            self.set_header("Content-Type", "application/json; charset=UTF-8")
            super(APIHandler, self).finish(chunk)

    def write_error(self, status_code, **kwargs):
        try:
            exc_info = kwargs.pop('exc_info')
            e = exc_info[1]

            if isinstance(e, errors.HTTPAPIError):
                pass
            elif isinstance(e, HTTPError):
                e = errors.HTTPAPIError(e.status_code, message=e.log_message)
            else:
                e = errors.HTTPAPIError(errors.ERROR_INTERNAL_SERVER_ERROR)

            exception = "".join([ln for ln
                                 in traceback.format_exception(*exc_info)])

            if status_code == errors.ERROR_INTERNAL_SERVER_ERROR \
                    and not options.debug:
                self.send_error_mail(exception)

            if options.debug:
                e.data["exception"] = exception
            logging.warning("Exception: %s", exception)

            self.clear()
            # always return 200 OK for API errors
            self.set_status(errors.HTTP_OK)
            self.set_header("Content-Type", "application/json; charset=UTF-8")
            self.finish(e)
        except Exception:
            logging.error(traceback.format_exc())
            return super(APIHandler, self).write_error(status_code, **kwargs)

    def send_error_mail(self, exception):
        """Override to implement custom error mail"""
        pass


class ErrorHandler(BaseHandler):

    """Default 404: Not Found handler."""

    def prepare(self):
        super(ErrorHandler, self).prepare()
        raise HTTPError(errors.ERROR_NOT_FOUND)


class APIErrorHandler(APIHandler):

    """Default API 404: Not Found handler."""

    def prepare(self):
        super(APIErrorHandler, self).prepare()
        raise errors.HTTPAPIError(errors.ERROR_NOT_FOUND)



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
from . import Stats, STATS_TIME_UNIT_DAY
from tornado.options import define, options

define("stats_time_expire", (72 + 1) * 3600, int,
       "expire time for stats records")
define("stats_db_sync_max_units", 5, int,
       "Max time units will be tried to synced")
define("stats_db_sync_min_views", 5, int,
       "Min views will be synced")


class PageviewStats(Stats):
    _stats_key_prefix = "stats:pv"

    def __init__(self, name, stats_redis, stats_db,
                 stats_time_unit=STATS_TIME_UNIT_DAY):
        self.stats_time_unit = stats_time_unit
        self.current_section = None
        super(PageviewStats, self).__init__(name, stats_redis, stats_db)

    def stats(self, message):
        """PV stats

        Message's keys:
        "page, timestamp" are required keys, "rid, view" are optional keys.
        """
        page = message["page"]
        timestamp = message["timestamp"]
        rid = message.get("rid", None)
        # optional params
        views = int(message.get("views", 1))
        section = int(timestamp) / self.stats_time_unit

        base_key = "%s:%s" % (self._stats_key_prefix, self.name)

        if rid:
            page_key = "%s-%s" % (page, rid)
        else:
            page_key = page

        pipe = self.redis.pipeline()
        # stats:api:time:2271792, records within time_unit
        time_key = "%s:time:%s" % (base_key, section)
        pipe.zincrby(time_key, page_key, views)
        pipe.expire(time_key, options.stats_time_expire)
        pipe.execute()

        if self.current_section != section:
            self.current_section = section
            self.prepare_to_sync()

    def sync_to_db(self):
        base_key = "%s:%s" % (self._stats_key_prefix, self.name)
        for i in xrange(1, options.stats_db_sync_max_units):
            section = self.current_section - i
            time_key = "%s:time:%s" % (base_key, section)
            count = self.redis.zcard(time_key)
            if count:
                limit = 200
                offset = 0
                while True:
                    logs = self.redis.zrange(time_key, offset,
                                             offset + limit - 1, withscores=True)
                    if logs:
                        self._sync_views_to_db(logs, section)
                        offset += len(logs)
                    else:
                        break

                self.redis.delete(time_key)

            i += 1

    def _sync_views_to_db(self, logs, section):
        viewed_at = section * self.stats_time_unit
        params = []
        for page_key, views in logs:
            v = page_key.split("--")
            if len(v) >= 2:
                rid = v[-1]
            else:
                rid = None
            page = v[0]
            if views > int(options.stats_db_sync_min_views):
                params.append((self.name, page, views, rid, viewed_at))
                logging.debug("sync_views_to_db: %s %s" % (page_key, int(views)))

        self.db.executemany("INSERT into pageview_logs (name, page, views, rid, "
                            "viewed_at) values (%s, %s, %s, %s, FROM_UNIXTIME(%s))",
                            params)



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
import functools
from tornado.escape import to_unicode
from djinn import errors


class Validator(object):

    """Basic Validator class"""
    _error_message = "%s is an invalid param"

    def __init__(self, param, default=None, message=None, code=400):
        self.param = param
        self.default = default
        self.message = message
        self.code = code

        def _validate(func):
            @functools.wraps(func)
            def __validate(handler, *args, **kwargs):
                # validate handler.arguments
                self.handler = handler
                self.validate()
                return func(handler, *args, **kwargs)

            return __validate
        self._validate = _validate

    def __call__(self, *args, **kwargs):
        return self._validate(*args, **kwargs)

    def validate(self):
        raise NotImplementedError()

    @property
    def get_argument(self):
        return self.handler.get_argument

    def handle_error(self, param, *args):
        _ = self.handler.locale.translate
        args_ = [_(param), ]
        args_.extend(args)
        args = tuple(args_)

        if not self.message:
            message = _(self._error_message) % args
        else:
            if "%s" in self.message:
                message = _(self.message) % args
            else:
                message = _(self.message)

        raise errors.HTTPError(self.code, message)


class Required(Validator):
    _error_message = "%s is required"

    def __init__(self, param, message=None, code=400):
        super(Required, self).__init__(param, None, message, code)

    def validate(self):
        value = self.get_argument(self.param, None)
        if not value:
            self.handle_error(self.param)


class String(Validator):
    _error_message = "%s's length is invalid(%s-%s)"

    def __init__(self, param, default="", message=None,
                 max_len=200, min_len=0, as_unicode=True, code=400):
        self.max_len = max_len
        self.min_len = min_len
        self.as_unicode = as_unicode
        super(String, self).__init__(param, default, message, code)

    def validate(self):
        value = self.get_argument(self.param, self.default)
        if self.as_unicode:
            length = len(to_unicode(value))
        else:
            length = len(value)

        if length < self.min_len or length > self.max_len:
            self.handle_error(self.param, self.min_len, self.max_len)


class PlainText(String):
    _error_message = "%s is an invalid plain text(a-z, A-Z,0-9 and _)"

    def __init__(self, param, default="", message=None,
                 max_len=200, min_len=0, code=400):
        super(PlainText, self).__init__(param,
                                        default, message, max_len, min_len, code)
        self.regex = re.compile('^[a-zA-Z0-9_]{%d,%d}$' %
                                (self.min_len, self.max_len), re.IGNORECASE)

    def validate(self):
        value = self.get_argument(self.param, self.default)
        if not self.regex.match(value):
            self.handle_error(self.param)

        super(PlainText, self).validate()


class Integer(Validator):
    _error_message = "%s should be integer"

    def __init__(self, param, default=0, message=None, code=400):
        super(Integer, self).__init__(param, default, message, code)

    def validate(self):
        try:
            int(self.get_argument(self.param, self.default))
        except Exception:
            self.handle_error(self.param)


class Enum(Validator):
    _error_message = "%s should be in %s"

    def __init__(self, param, type_=None, enum=None, default=None, message=None, code=400):
        self.enum = enum if enum else ()
        self.type_ = type_
        super(Enum, self).__init__(param, default, message, code)

    def validate(self):
        value = self.get_argument(self.param, self.default)

        if self.type_:
            try:
                value = self.type_(value)
            except TypeError:
                # ingore type error
                pass

        if value not in self.enum:
            # try type detection
            for e in self.enum:
                e_type = e.__class__
                try:
                    if e_type(value) == e:
                        break
                except:
                    continue
            else:
                self.handle_error(self.param, self.enum)


class Regex(Validator):
    _error_message = "%s is invalid"

    def __init__(self, param, regex, flags=re.IGNORECASE, default="",
                 message=None, code=400):
        self.regex = re.compile(regex, flags)
        super(Regex, self).__init__(param, default, message, code)

    def validate(self):
        value = self.get_argument(self.param, self.default)
        if not self.regex.match(value):
            self.handle_error(self.param)


class Email(Regex):
    _error_message = "%s is an invalid email address"

    def __init__(self, param, message=None, code=400):
        # borrow email re pattern from django
        regex =  r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*" \
            r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-011\013\014\016-\177])*"' \
            r")@(?:[A-Z0-9]+(?:-*[A-Z0-9]+)*\.)+[A-Z]{2,6}$"
        flags = re.IGNORECASE
        default = ""

        super(Email, self).__init__(param, regex,
                                    flags, default, message, code)


required = Required
string = String
integer = Integer
enum = Enum
regex = Regex
email = Email
plaintext = PlainText



# -*- coding: utf-8 -*-
#
# Copyright(c) 2013 lottouch.com
#
# @author: Felinx Lee <felinx.lee@gmail.com>
# Created on Dec 17, 2013
#

import time
import os
import traceback
import argparse
import logging
from tornado import escape
from tornado.options import define, options

STATS_TIME_UNIT_MINTUE = 60  # 1 minute
STATS_TIME_UNIT_HOUR = 3600  # 1 hour
STATS_TIME_UNIT_DAY = 24 * 3600  # 1 day
STATS_TIME_UNIT_WEEK = 7 * 24 * 3600  # 1 week

define("stats_db_sync_delay", 300, int, "Delay to make sure data complete")
define("stats_synclock_time", 600, int, "Lock time to finish sync")


class Stats(object):
    _stats_key_prefix = "stats"

    def __init__(self, name, stats_redis, stats_db):
        self.name = name
        self.redis = stats_redis
        self.db = stats_db
        self.message = None
        self._sync_timestamp = None

    def process(self, message):
        r = True
        try:
            self.stats(message)
        except Exception, e:
            logging.error(e)
            logging.error(traceback.format_exc())
            # result should be False for error processing
            r = False

        if r and self.ready_to_sync:
            try:
                lock = self.get_sync_lock()
                if lock:
                    self.sync_to_db()
                    self.release_sync_lock()
            except Exception, e:
                # ingore db sync error, just log it
                logging.error(e)
                logging.error(traceback.format_exc())

        return r

    def stats(self, message):
        raise NotImplemented()

    def sync_to_db(self):
        raise NotImplemented()

    @property
    def ready_to_sync(self):
        if self._sync_timestamp and \
            time.time() - self._sync_timestamp > options.stats_db_sync_delay:
            self._sync_timestamp = None
            return True
        else:
            return False

    def prepare_to_sync(self):
        self._sync_timestamp = time.time()

    def get_sync_lock(self):
        """Sync to db lock.

        To make sure only one worker to sync data to MySQL database.
        """
        key = "%s:%s:synclock" % (self._stats_key_prefix, self.name)
        r = self.redis.setnx(key, os.getpid())
        if r:
            self.redis.expire(key, options.stats_synclock_time)
            return True
        else:
            return False

    def release_sync_lock(self):
        self.redis.delete("%s:%s:synclock" % (self._stats_key_prefix,
                                              self.name))
        self._sync_to_db = False



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import redis
from ..errors import DatastoreError

manager = None


def setup(redis_pool):
    global manager

    if manager is None:
        manager = RstoreManager(redis_pool)
    return manager


class RstoreManager(object):
    _datastore_pool = {}

    def __init__(self, datastore_pool):
        for k, v in datastore_pool.iteritems():
            RstoreManager._datastore_pool[k] = redis.Redis(v[0], v[1])

    def __getattr__(self, instance):
        r = self._datastore_pool.get(instance, None)
        if not r:
            raise DatastoreError("Redis %s instance does not exist"
                                 % instance)

        return r


class RedistoreException(Exception):
    pass



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tornado import escape
from tornado.web import HTTPError

# HTTP status code
HTTP_OK = 200
ERROR_BAD_REQUEST = 400
ERROR_UNAUTHORIZED = 401
ERROR_FORBIDDEN = 403
ERROR_NOT_FOUND = 404
ERROR_METHOD_NOT_ALLOWED = 405
ERROR_INTERNAL_SERVER_ERROR = 500
# Custom error code
ERROR_WARNING = 1001
ERROR_DEPRECATED = 1002
ERROR_MAINTAINING = 1003
ERROR_UNKNOWN_ERROR = 9999


class DjinnError(Exception):
    pass


class DatastoreError(DjinnError):
    pass


class TemplateContextError(DjinnError):

    """Template context variable does not exist."""
    pass


class HTTPAPIError(HTTPError):

    """API error handling exception

    API server always returns formatted JSON to client even there is
    an internal server error.
    """

    def __init__(self, status_code=ERROR_UNKNOWN_ERROR, message=None,
                 error=None, data=None, *args, **kwargs):
        assert isinstance(data, dict) or data is None
        message = message if message else ""
        assert isinstance(message, basestring)

        super(HTTPAPIError, self).__init__(int(status_code),
                                           log_message=message, *args, **kwargs)

        self.error = error if error else \
            _error_types.get(self.status_code, _unknow_error)
        self.message = message if message else \
            _error_messages.get(self.status_code, _unknow_message)
        self.data = data if data is not None else {}

    def __str__(self):
        err = {"meta": {"code": self.status_code, "error": self.error}}

        if self.data:
            err["data"] = self.data

        if self.message:
            err["meta"]["message"] = self.message % self.args

        return escape.json_encode(err)


# default errors
_unknow_error = "unknow_error"
_unknow_message = "未知错误,请稍后再试"
_error_types = {400: "bad_request",
                401: "unauthorized",
                403: "forbidden",
                404: "not_found",
                405: "method_not_allowed",
                500: "internal_server_error",
                1001: "warning",
                1002: "deprecated",
                1003: "maintaining",
                9999: _unknow_error}

_error_messages = {400: "请求参数错误",
                   401: "用户未登录",
                   403: "您未登录或无权访问",
                   404: "资源不存在",
                   405: "客户端请求方式错误",
                   500: "服务器这会开了点小差，程序猿正加紧修复中，请稍后再试",
                   1001: "错误警告",
                   1002: "接口不兼容，请升级版本",
                   1003: "服务器维护中，请稍后再试",
                   9999: _unknow_error}



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import distutils.core
try:
    import setuptools
except ImportError:
    pass

version = "1.0.6"

distutils.core.setup(
    name="djinn",
    version=version,
    packages=["djinn", "djinn.datastore", "djinn.stats"],
    package_dir={'djinn': 'djinn'},
    package_data={'djinn': ['translations/*.csv']},
    author="Felinx Lee",
    author_email="felinx.lee@gmail.com",
    url="https://github.com/palmhold/djinn",
    license="http://www.apache.org/licenses/LICENSE-2.0",
    description="Djinn is a micro-framework which wraps Tornado to make it easy to write a web application using Tornado.",
)



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
from tornado.web import HTTPError


def authenticated(method):
    """Decorate methods with this to require that the user be logged in.

    Just raise 401
    """
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self.current_user:
            raise HTTPError(401)
        return method(self, *args, **kwargs)
    return wrapper



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time
import logging
import random
from torndb import Connection as BaseConnection
from tornado.options import options
from tornado import escape

from ..errors import DatastoreError

manager = None


def setup(datastore_pool):
    global manager

    if manager is None:
        manager = MysqlManager(datastore_pool)
    return manager


class MysqlManager(object):
    _datastore_pool = {}

    def __init__(self, datastore_pool):
        for k, v in datastore_pool.iteritems():
            MysqlManager._datastore_pool[k] = MysqlMSConnection(v[0], v[1])

    def __getattr__(self, instance):
        r = self._datastore_pool.get(instance, None)
        if not r:
            raise DatastoreError("Mysql %s instance does not exist"
                                 % instance)

        return r


class MysqlMSConnection(object):

    """Mysql master & slave connection

    Manage datastore connection instances, it selects master or slave instance
    according to SQL statement automatically.

    """

    def __init__(self, master, slaves):
        self.master = Connection(master)

        self._slave_conns = []
        for slave in slaves:
            self._slave_conns.append(Connection(slave))

    @property
    def query(self):
        return self.slave.query

    @property
    def get(self):
        return self.slave.get

    @property
    def execute(self):
        return self.master.execute

    @property
    def executemany(self):
        return self.master.executemany

    @property
    def insert(self):
        return self.slave.insert

    @property
    def update(self):
        return self.slave.update

    @property
    def slave(self):
        return random.choice(self._slave_conns)


class Connection(BaseConnection):

    def __init__(self, options):
        default_options = {
            "host": "localhost:3306",
            "database": "test",
            "user": None,
            "password": None,
            "charset": "utf8mb4",
            "max_idle_time": 7 * 3600,
            "connect_timeout": 0,
            "time_zone": "+8:00"
        }
        default_options.update(options)

        super(Connection, self).__init__(**default_options)

    def _execute(self, cursor, query, parameters, kwparameters):
        # Override default _execute to log executing info when log_query is on
        parameters_ = []
        for parameter in parameters:
            if isinstance(parameter, unicode):
                parameter = escape.utf8(parameter)
            parameters_.append(parameter)
        parameters = tuple(parameters_)

        if options.LOG_DB_QUERY:
            sql = query % parameters
            try:
                start = time.time()
                r = super(Connection, self)._execute(cursor, query, parameters,
                                                     kwparameters)
                elapse = time.time() - start
                logging.info("SQL executing elapse %s seconds on %s: %s",
                             elapse, self.host, sql)

                return r
            except Exception, e:
                logging.error("SQL: %s", sql)
                raise e
        else:
            return super(Connection, self)._execute(cursor, query, parameters, kwparameters)
